#include "std.h"
#include "datamove/datapoint.h"
#include "datamove/datahandle.h"
#include "misc/log_time.h"

extern "C" {
int ngremove(int argc,char** argv) {
  LogTime::active=false;
  bool verbose = false;
  bool errcont = false;
  for(;;) {
    opterr=0;
    int optc=getopt(argc,argv,"+hvd:c");
    if(optc == -1) break;
    switch(optc) {
      case 'h': {
        olog<<"ngremove [-h] [-v] [-d level] [-c] url"<<endl;
        olog<<" -h - print help"<<endl;
        olog<<" -v - print version"<<endl;
        olog<<" -d - set debug level"<<endl;
        olog<<" -c - continue removing meta information even if failed to delete physical file instance"<<endl;
        exit(1);
      }; break;
      case 'v': {
        olog<<"ngremove: version "<<VERSION<<endl;
        exit(0);
      };
      case 'd': {
        char* p;
        int i = strtol(optarg,&p,10);
        if(((*p) != 0) || (i<0)) {
          olog<<"Improper debug level '"<<optarg<<"'"<<endl;
          exit(1);
        };
        verbose=true;
        LogTime::level=i;
      }; break;
      case 'c': {
        errcont=true;
      }; break;
      case '?': {
        olog<<"Unsupported option '"<<(char)optopt<<"'"<<endl;
        exit(1);
      }; break;
      case ':': {
        olog<<"Missing parameter for option '"<<(char)optopt<<"'"<<endl;
        exit(1);
      }; break;
      default: {
        olog<<"Undefined processing error"<<endl;
        exit(1);
      };
    };
  };
  char* file_url = argv[optind+0];
  if(file_url == NULL) {
    olog<<"Missing url" << endl; return 1;
  };
  if(argv[optind+1] != NULL) {
    olog<<"Unrecognized option: "<<argv[optind+1]<<endl; return 1;
  };
  DataPoint url(file_url);
  if(!url) {
    olog<<"Unsupported url"<<endl; return 1;
  };
  DataHandle handle(&url);
  handle.secure(false);
  bool remove_lfn = true;
  if(url.have_locations()) { /* pfn or plain url */
    remove_lfn=false;
  };
  if(!url.meta_resolve(true)) {
    if(remove_lfn) {  
      odlog(0)<<"No locations found - probably no more physical instances"<<endl;
    };
  };
  /* go through all locations and try and remove files
     physically and their metadata */
  for(;url.have_location();) {
    if(verbose) odlog(0)<<"Removing "<<url.current_location()<<endl;
    if(!handle.remove()) {
      odlog(0)<<"Failed to delete physical file"<<endl;
      if(!errcont) { 
        url.next_location(); continue;
      };
    };
    if(!url.meta()) {
      url.remove_location();
    }
    else {
      if(verbose) odlog(0)<<"Removing metadata in "<<url.current_meta_location()<<endl;
      if(!url.meta_unregister(false)) {
        odlog(0)<<"Failed to delete meta-information"<<endl;
        url.next_location();
      }
      else { url.remove_location(); };
    };
  };
  if(url.have_locations()) return 1;
  if(url.meta()) {
    if(remove_lfn) {
      if(verbose) odlog(0)<<"Removing logical file from metadata "<<url.canonic_url()<<endl;
      if(!url.meta_unregister(true)) {
        odlog(0)<<"Failed to delete logical file"<<endl;
        return 1;
      };
    };
  };
  if(verbose) { cout<<"Remove succeeded"<<endl; };
  return 0;
}
};

